/*******************************************************************************
* FILE NAME: pid.c
*
* DESCRIPTION: 
*  This is a set of sample functions that implement a simple PID 
*  control algorithm to control the position or velocity of a motor if we 
*  have an encoder onboard.
*
* USAGE:
*  Place the function pid() in the user_routines.c function
*  Process_Data_From_Master_uP() for each motor that we want
*  to control.
*******************************************************************************/
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "printf_lib.h"
#include "encoder.h"
#include "pid.h"


/* Globals */
long int pid_time = 1;
struct pid_data motor_info[2] = {{PID_STOP,0,0,0,0,0,0,0,0,0,0,0},
                                 {PID_STOP,0,0,0,0,0,0,0,0,0,0,0}};

/* Put this in User_Initialization */
void pid_init(void)
{
wheel_l = PWM_ZERO + motor_info[LEFT].pwm;
wheel_r = PWM_ZERO + motor_info[RIGHT].pwm;
}

/* Call this each cycle for each motor */
void pid(int motor)
{
long int pos_last = motor_info[motor].pos;
long int pos_error_last = motor_info[motor].pos_error;
long int vel_last = motor_info[motor].vel;
long int vel_error_last = motor_info[motor].vel_error;


motor_info[motor].pos = Get_Encoder_Count(motor);

if (PID_POS == motor_info[motor].pid_mode)
  {
  /* Position control */
  motor_info[motor].pos_error = motor_info[motor].pos_cmd - 
                                motor_info[motor].pos;
  motor_info[motor].pos_error_i += motor_info[motor].pos_error;
  motor_info[motor].pos_error_d = motor_info[motor].pos_error - 
                                  pos_error_last;
  motor_info[motor].pwm = (KP_P * motor_info[motor].pos_error) + 
                          ((KI_P * motor_info[motor].pos_error_i)/pid_time) + 
                          (KD_P * motor_info[motor].pos_error_d);
  motor_info[motor].pwm /= DIV_P;
  }
else if (PID_VEL == motor_info[motor].pid_mode)
  {
  /* Velocity Control */
  motor_info[motor].vel = ((vel_last * 9) + (motor_info[motor].pos - pos_last) * HZ)/10;
  motor_info[motor].vel_error = motor_info[motor].vel_cmd -
                                motor_info[motor].vel;
  motor_info[motor].vel_error_i += motor_info[motor].vel_error;
  motor_info[motor].vel_error_d = motor_info[motor].vel_error -
                                vel_error_last;

  /* Velocity Control */
  motor_info[motor].pwm = (KP_V * motor_info[motor].vel_error) + 
                          ((KI_V * motor_info[motor].vel_error_i)/pid_time) + 
                          (KD_V * motor_info[motor].vel_error_d);
  motor_info[motor].pwm /= DIV_V;
  }
else  /* STOP */
  {
  motor_info[motor].pwm = 0;
  }

/* Compensate for deadband */
if  (motor_info[motor].pwm > 0)
  motor_info[motor].pwm += DEADBAND;
else if (motor_info[motor].pwm < 0)
  motor_info[motor].pwm -= DEADBAND;

/* limit pwm command */
if (motor_info[motor].pwm > (MAX_PWM-PWM_ZERO))
  {
  motor_info[motor].pwm = (MAX_PWM-PWM_ZERO);
  }
else if (motor_info[motor].pwm < (MIN_PWM-PWM_ZERO))
  {
  motor_info[motor].pwm = (MIN_PWM-PWM_ZERO);
  }

#ifdef DEBUG

if (pid_time%20 == 0) 
  {
  printf("vel[%d] = %d   pos = %d pwm = %d\n", (int) motor, 
                           (int) motor_info[motor].vel, 
                           (int) motor_info[motor].pos,
                           (int) motor_info[motor].pwm);
  
  }
#endif

switch (motor)
  {
  case LEFT:
    {
    wheel_l = (unsigned char)(PWM_ZERO + LEFT_DIR * motor_info[LEFT].pwm);
	break;
    }
  case RIGHT:
    {
    wheel_r = (unsigned char)(PWM_ZERO + RIGHT_DIR * motor_info[RIGHT].pwm);
	break;
    }
  default:
	{
	printf("Unknown motor");
	break;
	}
  }
}

void set_pos(int motor, long int pos)  // pos = encoder counts
                                       // 256 = 1 rev
{
motor_info[motor].pid_mode = PID_POS;
motor_info[motor].pos_cmd = pos;
pid_time = 1;     // Start over
}
 
void set_vel(int motor, int vel)  // vel = encoder counts / sec
{                                 // 256 = 1 rev/sec
motor_info[motor].pid_mode = PID_VEL;
motor_info[motor].vel_cmd = vel;
pid_time = 1;     // Start over
}


void set_pid_stop(int motor)
{
motor_info[motor].pid_mode = PID_STOP;
motor_info[motor].pwm = 0;
Set_Encoder_Count(motor, 0L);
motor_info[motor].pos_cmd = 0;
motor_info[motor].pos_error = 0;
motor_info[motor].pos_error_i = 0;
motor_info[motor].pos_error_d = 0;
motor_info[motor].vel_cmd = 0;
motor_info[motor].vel_error = 0;
motor_info[motor].vel_error_i = 0;
motor_info[motor].vel_error_d = 0;
}

long int get_pos(int motor)  // pos = encoder counts
{                               
return motor_info[motor].pos;
}

long int get_vel(int motor)  // vel = encoder counts / sec
{                               
return motor_info[motor].vel;
}

long int Get_Encoder_Count(int motor)
{
if (motor == 0)
  return Get_Left_Encoder_Count();
else if (motor == 1)
  return Get_Right_Encoder_Count();
else 
  return 0;
}

void Set_Encoder_Count(int motor, long int val)
{
if (motor == 0)
  Set_Left_Encoder_Count(val);
else if (motor == 1)
  Set_Right_Encoder_Count(val);

return;
}
