/*******************************************************************************
* FILE NAME: pid.h
*
* DESCRIPTION: 
*  Header file for pid.c
*  This is a set of sample functions that implement a simple PID 
*  control algorithm to control the position and velocity of a 
*  motor if we have an encoder onboard.
*
* USAGE:
*  Place the function pid() in the user_routines.c function
*  Process_Data_From_Master_uP() for each motor that we want
*  to control.  In this code, it is in robot_command() in
*  robot.c
*******************************************************************************/

/******************************************************************************
                 What is PID Control ?

PID control can be used any time you have a variable that you want
to control and a way of measuring that variable.

P - Proportional - The amount of control is based on (proportional to) the 
                   error between where we want to be and where we are.  

I - Integral - The amount of control is based on the total error over
               time - the summation or integral

D - Differential - The amount of control is based on how fast the error
                   is changing - the difference.

*******************************************************************************/
/******************************************************************************

                 What are Gains ?

The gains tell you how important each of the PID terms are.

Kp - Proportional Gain - Determines how fast your system reacts to error

Ki - Integral Gain - Determines how hard your system will push to overcome
                     error.  

Kd - Differential Gain - Limits the change in response to error.  Helps to
                         dampen or smooth the reactions. 
*******************************************************************************/
/******************************************************************************

                  How do I tune my system?

Start by setting the gains low.  You can set Ki and Kd to zero to start.

Increase Kp until the system starts to react quickly enough.  It will 
overshoot if you set it too high.

Now increase Kd to compensate for overshoot.  Now the system should react
smoothly.

But you might notice that it never reaches the goal.  That is because 
resistance in the system is holding it back and as you near the goal, the 
proportional term gets smaller.  Now it is time to increase Ki.  Over time
the error will build and the I term allows the system to overcome resistance.

Now go back and adjust each of the terms to get the response you want.
*******************************************************************************/
#ifndef __pid_h_
#define __pid_h_

/* We will assume all the motors have the same gains */

  /***********/
 /* Defines */
/***********/

/* These are the gains for position control (when you want the 
   motor to maintain a certain position). The first is the 
   proportional gain, the second is the integral gain and the 
   third is the differential gain.  The last parameter is a 
   divisor used to scale the output to the correct level. */
#define KP_P (5)
#define KI_P (0)
#define KD_P (50)
#define DIV_P (50)

/* These are the gains for velocity control (when you want the 
   motor to maintain a certain velocity). The first is the 
   proportional gain, the second is the integral gain and the 
   third is the differential gain.  The last parameter is a 
   divisor used to scale the output to the correct level. */
#define KP_V (10)
#define KI_V (1)
#define KD_V (20)
#define DIV_V (200)

/* PID modes */
#define PID_POS    1
#define PID_VEL    2
#define PID_STOP   3

/* Motor defines */
#define LEFT  0
#define RIGHT 1

/* These define the motor control settings.  
   The max and min values are the greatest and smallest pwm
   values that you want to use on your motors. */
#define MAX_PWM 182
#define MIN_PWM 82
/* This is the center of the motor's deadband (the range
   of values wher it does not move). */
#define PWM_ZERO 132
/* Deadband is the number of steps above or below PWM_ZERO where
   the motor just starts to move. */   
#define DEADBAND 7

/* Define your pwm values for the left and right wheels. */
#define wheel_l pwm01
#define wheel_r pwm03
/* Try this if you have two motors on each side. */
//#define wheel_l pwm01 = pwm02
//#define wheel_r pwm03 = pwm04

/* Set these to change the motor direction,
   1 = forward, -1 = backward.  */
#define LEFT_DIR (-1)
#define RIGHT_DIR (1)

/* 26msec per cycle = 39 HZ */
#define HZ   39

  /***********/
 /* Structs */
/***********/

/* This structure holds the position and velocity data
   for each motor. */
struct pid_data 
{
	int pid_mode;
	long int pos;
	long int pos_cmd;
	long int pos_error;
	long int pos_error_i;
	long int pos_error_d;
	long int vel;
	long int vel_cmd;
	long int vel_error;
	long int vel_error_i;
	long int vel_error_d;
	long int pwm;
};

  /***********/
 /* Globals */
/***********/

/* Just increment the time each cycle.  This is used for the 
   integral term. */
extern long int pid_time;
/* This is the struct that stores the motor info between cycles */
extern struct pid_data motor_info[];

  /*************/
 /* Functions */
/*************/

/* These are the interface functions for PID control */
void pid_init(void);               // Put this in User_Initialization
void pid(int motor);               // Call this each cycle for each
                                   // motor

/* These are the API functions for low level motor control */
void set_pos(int motor, long int pos);  // pos = encoder counts
                                      
void set_vel(int motor, int vel);  // vel = encoder counts / sec
                                   
void set_pid_stop(int motor);      // Stop PID Control

long int get_pos(int motor);       // Get the current encoder position

long int get_vel(int motor);       // Get the encoder velocity

/* These are used internally by PID to interface with the 
   encoder code. */
long int Get_Encoder_Count(int motor);

void Set_Encoder_Count(int motor, long int val);

#endif /* __pid_h_ */
